//----------------------------------------------------------------------------
// SVG Viewer - Version 1.0
// A library of SVG renderer
// Copyright (C) 2008 Witold Jaworski
// Contact:	witold.jaworski@airplanes3d.net
//			http://airplanes3d.net
// 
// Contains elements from:
//			AGG - a graphic library (Copyright Maxim Shemanarev, mcseem@antigrain.com)
//          http://antigrain.com
// Uses functions from:
//          Expat - a XML parser library (Copyright James Clark)
// 
// SVG Viewer is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// SVG Viewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with SVG Viewer; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
// MA 02110-1301, USA.
//----------------------------------------------------------------------------

#ifndef SVG_ELEMENT_INCLUDED
#include "svg_parser.h"
#include "svg_request.h"
#define SVG_ELEMENT_INCLUDED

/*
Some general ideas about the SVG graph implementation.
SVG graph in the memory is mimicking up to some extent the root structure of
SVG XML document. There is a root node, which contains the first node, matching 
the <svg> element. Then, each node can have zero or more children. All the nodes
are instances of the classes inherited from the element class, defined here.

Because an XML element that defines a node of SVG graph has most of its
attributes optional, I have decided to store them not as the fields of the
classes, but in an uniform storage, maintained by the root node. Each node 
(element) has to define (store) its attributes during initialization.
They are ultimately stored in an array managed by the root, and the node receives
back an integer handle - the index of this attribute in the main storage.
Using a function get_attribute() it can retrieve this attribute value back.
Each element should define its attributes in the constructor, only!
Only this way it can be assumed, that they form a continous range, and all attributes
having handles from m_attribute_first to m_attribute_last belongs to this node. 
Because implementation of some methods (like element::get_attribute()) is based 
on this assumption, you must follow this rule.

Similar "handle" mechanism is used to preserve the node element itself. 
They are stored in the root storage. They are also available by the get_element() 
function. 

I have used this design pattern for te nodes, because some of them can reference 
the others (like textPath element - it references a path). What is not trivial
- the referenced element can be not loaded, yet, when the node that references it
is already commited to life (i.e. element::commit() method was already called). 

To resolve this problem, I have decided to utilize a "handle" approach. When something 
is needed to be referenced - it can be referenced, as a kind of a "placeholder" proxy
object. This is the reference struct, (see svg_parser.h). The "client" stores its handle
and does not need anything more. During further processing the content stored under this 
handle - the reference::elm field - will be replaced with the "real" object, but it does 
not bother the "client" node. Of course, there should be no placeholders at the end of 
the processing. Each of them is, on the final stage, the evidence of an unresolved 
reference, which means an error. 
*/

namespace svg
{
//////////////////////////////////////////////////////////////////////
// forward class declarations
//////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////
// element: base class for all elements of the SVG graph
//			(representation of the SVG document content in the RAM)
//////////////////////////////////////////////////////////////////////
class element  
{
	public:
		//////////////////////////////////////////////////////////////////////
		// Methods used for the creation (and destruction)
		//////////////////////////////////////////////////////////////////////
		//
		//Because of the Expat architecture, each element is constructed by
		//many subsequent calls to the following methods: 
		//	1. constructor,				- when a new XML tag was found 
		//	2. on_attribute()			- inside read_attributes: a call for each 
		//								  attribute found in the XML element.
		//								  It is guaranteed that this method will be 
		//								  always called to handle the "id" attribute.
		//								  Even when the xml element has no "id" attribute,
		//								  the constructor will call this method,
		//								  passing an empty string ("") as the value.
		//								  This way the element has the chance to presist
		//								  itself.
		//	3. optionally: on_content()	- when this tag has a textual content
		//	4. commit().				- when the end of this XML tag was found
		//
		//	Only these functions, and some helper ones which can be called inside 
		//	them, will receive the svg::parser object reference.
		//  You should not store this parser object as any pointer - it is an object
		//	created on the stack of the main function, and will be destroyed after
		//	return from the last svg::element::commit() call.
		//	
		//	The first call to the element when whole SVG graph was loaded, is
		//	the element::on_request(request::initialize). You can override the 
		//	this function to do some initialization there.
		//	After this call some other (like request::measure) calls
		//	are obtained by the element, until the final request::render call.
		//////////////////////////////////////////////////////////////////////
		//////////////////////////////////////////////////////////////////////

		//--------------------------------------------------------------------

		//Constructor (called inside Expat's start_element() callback function):
		//Arguments:
		//	parent:	the container element (container matches the xml container
		//			tag, but there may be a difference)
		//	level:	the level in the XML document hierarchy - helper information,
		//			used when reading the content
		//BEWARE: virtual methods of the child class are available AFTER its 
		//constructor call! do not place any special logic in the constructor!
		element(element *parent, unsigned level)
		{
			m_id = m_id_last = invalid_handle;
			m_attribute_first = m_attribute_last = invalid_handle;
			m_forward_use = invalid_handle;
			m_parent = parent;
			m_level = level;
			m_name = NULL;
			m_commited = 0;
		}

		//Called just after the element constructor: normally it should analyze and -
		//optionally - store this attribute
		//Arguments:
		//	prs:	parser, that actually prepares the whole thing
		//	name:	name of the attribute;
		//	value:	value of the attribute (text);
		//It is guaranteed, that this function will be called with the attribute
		//name = "id". When the source XML element no such attribute - the value = "".
		virtual void on_attribute(parser &prs, const char * name, const char *value);

		//Additional function, called after the constructor, and attributes when the 
		//corresponding XML element contains some inner character data (like a text, 
		//for example). Called inside Expat's oncontent() callback function
		//Arguments:
		//	prs:	parser, that actually prepares the whole thing
		//	s:		text, that was read
		//	len:	length of this text
		virtual void on_content(parser &prs, const char* s, int len) {}

		//Additional function, called after the constructor, arguments and content,
		//when Expat has reached the XML closing tag.
		//Commits successful creation of a new object.
		//Arguments:
		//	prs:	parser, that actually prepares the whole thing
		//It is used by Expat's end_element() callback function, or - in some cases -
		//by elements, that artifically create another elements without the xml source
		//(see svg::text - it creates helper svg::tspan objects). commit() is normally 
		//called only once in element's "life". In some situations (for example - inner 
		//svg::tspan in the svg::text element) it can be called more times. 
		virtual void commit(parser &prs) 
		{
			m_id_last = (prs.elements() -1 );
			m_commited ++;
		}

		//Returns number of the commit() calls 
		//(sometimes it is useful - for example in svg::text, for internal tspans)
		unsigned commited() {return m_commited;}

		//Destructor: nothing special
		virtual ~element() {}

	protected:
		char		*m_name;			//xml id of this element - 
										//usually points to containing reference::name buffer
										//but can point to anything else in the inherited
										//classes

		//All attributes of all SVG graph elements are stored in a common storage.
		//This storage is maintained by the svg::root element.
		//It was not reasonable to keep a copy of all possible attributes as a
		//class fields. There are too many of possibilites, especially among the
		//style attributes. Instead of "classic" approach, I have choosed the common
		//storage model. The element keeps the attribute handles (ids), only.
		//Their values are stored in one, continous attribute storage (agg::bvector) 
		//What's more, because of the specific SVG XML structure and its parsing algorithm,
		//all attributes of a particular element are stored in a continous sequence of 
		//ascending ids. This situation allows to keep in the element just the id of the 
		//first and the last attribute, keeping its size small.
		//Access to the attribute storage is given through the svg::parser and 
		//svg::request objects and their descended classes.
		handle		m_attribute_first;	//id of the first attribute
		handle		m_attribute_last;	//id of the last attribute

		//There is an exception from the rule of "continous attribute range" that belongs
		//to single element: there are some helper elements in the SVG graph, that have no
		//counterparts in the SVG XML source. For example: the svg::dash element,
		//which contains the spans of the line pattern definition. The span of its 
		//attributes lies, in fact, between the m_attribute_first and m_attribute_last
		//of its parent. Because svg::dash uses just one kind of the attributes -
		//- attribute::dash_span - as long as they are ignored (not used) in further
		//processing by the parent, it makes no harm to the overall idea.

		//SVG elements are the only elements in the parser, that are created
		//on the heap, using the new operator. To control their "life and death",
		//their addresses are still contained in one, common storage. This is storage
		//of svg::reference structures. It allows svg::elements to bind into another
		//svg::element (a text to a path, for example) even before the target was found
		//in the source file and prepared! It is possible, because svg:element does
		//not keep directly the addresses of the other elements. It keeps just a handle
		//- the id of the other element in the common storage. 
		//Access to the element storage is given through the svg::parser and 
		//svg::request objects and their descended classes.

		//Another aspect of forward references: There are situations when an element
		//references another BEFORE it was read from the XML source and created
		//matching svg::element in the SVG graph. Usually it happens through "xlink:href"
		//XML attribute.
		//For sequential reading of the XML file, which is applied here, it is a problem.
		//To resolve it, svg::parser creates a "place" for referenced element just after 
		//the element that contains the "xlink:href" attribute. At the begining It 
		//contains just the name. Each newly read from the XML source element is 
		//chcecked against this name. When it match - newly created svg::element fills 
		//the "place" prepared before, instead of appedning to the end of element list.
		//It creates some problems, because, in this way, the referenced element is not
		//on the proper place, and, for example, it can be rendered not according the
		//order implied by the XML source.
		//To resolve this problem:
		// -	in place where the referenced element should properly appear, an unnamed 
		//		svg::use is placed, pointing to the original data. It behaves like a 
		//		"proxy". It transfers the requests, it receives, to the referenced 
		//		element, and responds back with its "answers"
		// -	simulatelously, the original data element receives a handle to this 
		//		svg::use - as a mark, that all undirect requests from other elements 
		//		should be ignored. (This way, from the view of the "neighbour" elements
		//		in the storage, it is a "dumb" element, that does nothing).
		//See parser::store for details of creation such svg::use element. There is a 
		//special constructor used, which creates thsi svg::use with a special flag - 
		//m_from_xml - cleared. This flag forces a different behaviour with incoming 
		//requests, comparing to the case of an <use> element, created from XML data.

		handle		m_id;			//id of this element (when registered)
		handle		m_forward_use;  //valid, when element has been placed as a data
									//that is used elsewhere AFTER the original place.
									//(Used for the elements referenced in SVG graph
									//before they were loaded - see explanation above)
		handle		m_id_last;		//id of the element - last child of this element
									//(in fact, it can be one child of the children)
									//but, for sure, the last descendent of this element

		//returns the first child id (for iteration)
		handle first_child()
		{
			if (elements() > 0) 
			{
				//Sometimes, when the element was referenced before definition
				//number of the elements is LOWER than difference between m_id and m_id_last:
				return m_id_last - elements() + 1;
			}
			else
			{
				if(is_valid(m_id_last)) 
				{
					return m_id_last+1;
				}
				else
				{
					return invalid_handle;
				}
			}
		}
		//In this functions parser changes the m_attribute_last and
		//(only once, if initialized) initializes the m_attribute_first;
		friend void parser::store(element &elm, attribute attr);

	public:
		//////////////////////////////////////////////////////////////////////
		// Methods used for request handling
		// (Requests is just a message, "bubbling" through the SVG graph tree)
		// It is the basic way to do anything with the SVG graph, once it has
		// been loaded. Request are usually sent by the root.
		//////////////////////////////////////////////////////////////////////
		//Forwards the request to the children on the SVG graph
		//Arguments:
		//	req:	an instance of one of classes that inherits from request 
		//			class.
		//Returns the first id of not processed SVG graph element
		//This function is created to make possible the recurrent walk around
		//the SVG graph.
		handle forward(request &req) 
		{	
			if (req.can_forward())
			{
				handle id = first_child();
				 
				while (id <= m_id_last)
				{
					element *elm =	req.get_element(id);

					//valid forward_use(): sometimes among child elements in the resulting
					//graph  you will find an element, which was not defined in the source XML.
					if(elm && !is_valid(elm->forward_use())) 
					{
						id = elm->on_request(req); 
					}
					else //sometimes (missing font?) element can be null...
					{
						id++;
					}//end if
				}//end while

				return id;
			}//end if
			else return skip(req);
		}
		//alternative for the forward function: skips the processing of given
		//request on all the descendants in the SVG graph.
		//Arguments:
		//	req:	an instance of one of classes that inherits from request 
		//			class. NOT USED. I have placed it here just because of
		//			"symmetrical" look in the code, that uses them!
		//Returns the first id of not processed SVG graph element
		handle skip(request &req) {return m_id_last + 1;}

		//Directs the request to the root on the SVG graph (it means 
		//opposite direction, than in the forward() function.
		//All these request should have return can_forward() = false
		//Arguments:
		//	req:	an instance of one of classes that inherits from request 
		//			class.
		void to_root(request &req) 
		{	
			//only non-forwardable requests are eligible for the 
			//root communication! (just do not send any typical kind of requests,
			//like "render")
			if (!req.can_forward()) 
			{
				if (parent()) parent()->to_root(req);
				on_request(req); 
			}
		}

		//Handles the request, and forwards it to the descendants.
		//Arguments:
		//	req:	the request instance - usually with helper data.
		//			Use the req.type() to deterimine, what is it,
		//			and what class it should be cated to.
		//Returns the first id of not rendered SVG graph element
		//Default implementation just forwards the request down the SVG graph.
		//Rememeber, that each "return" in the overrides should return the
		//result of forward() or skip() functions! Otherwise you can have some
		//problems!
		virtual handle on_request(request &req);
		
		//////////////////////////////////////////////////////////////////////
		// Helper properties, used for loading from an XML file:
		//////////////////////////////////////////////////////////////////////
		//returns the parent element (NULL for the root node)
		element *parent() {return m_parent;}

		//returns the xml tag name of the element (sometimes it may be useful)
		// -  pure virtual, to override in the ultimate classes
		virtual const char * tag() = 0;

		//returns a name (the value of the xml "id" attribute) of an element.
		//Names longer than 31 characters are truncated to 31 letters
		const char *name() {return m_name;}

		//a rudimentary kind of runtime type-checking
		//Arguments:
		//	cls:	a name of an elment family, like "shape", "text", etc.
		//Returns true, when this instance can be casted into requested class
		virtual bool is_of(const char *cls) 
		{
			return (strcmp(tag(),cls) == 0);
		}


		//returns the global id of this element in the SVG scene graph
		//If not registered globally, returns svg::invalid_handle value
		handle id() {return m_id;}

		//returns number of the elements contained inside
		unsigned elements() 
		{
			if (is_valid(m_id)) 
			{
				if(is_valid(m_forward_use)) //sometimes element can be placed 
				{						    //ahead of its definition:	
					return m_id_last - m_forward_use; 
				}
				else
				{
					return m_id_last - m_id;
				}
			}
			else
			{
				return 0;
			}
		}

		//Returns id of first occurence an attribute that has given type
		//Arguments:
		//	req:			actual request object, that manages the attribute 
		//					storage
		//	type:			the type of the attribute
		//	search_from:	optional - the handle id that should be used
		//					Usin this parameter, you can find another
		//					occurrence of an attribute of given type.
		//Beware: it is not optimized function - it performs linear search.
		//Use it during preparation of the SVG graph, not during renderings!
		handle get_id(request &req, attribute::kind type, 
											handle search_from = invalid_handle);

		//returns the level of this element in the XML hierarchy - information
		//needed during reading the XML source file.
		unsigned level() {return m_level;}

		handle forward_use() {return m_forward_use;}

	private:
		element		*m_parent;  //container element - NULL for the root node  
		unsigned	 m_level;	//nest level of the XML element (stored here for the XML parsing algorithm)
		unsigned	 m_commited;//count of the commit() calls
		//In this functions parser assigns the m_id and m_name, 
		//optionally sets the m_forward_use field to cref.
		friend void parser::assign(handle id, element *elm, handle cref);
}; //element

}; //namespace svg
#endif //SVG_ELEMENT_INCLUDED

