//----------------------------------------------------------------------------
// SVG Viewer - Version 1.0
// A library of SVG renderer
// Copyright (C) 2008 Witold Jaworski
// Contact:	witold.jaworski@airplanes3d.net
//			http://airplanes3d.net
// 
// Contains elements from:
//			AGG - a graphic library (Copyright Maxim Shemanarev, mcseem@antigrain.com)
//          http://antigrain.com
// Uses functions from:
//          Expat - a XML parser library (Copyright James Clark)
// 
// SVG Viewer is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// SVG Viewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with SVG Viewer; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
// MA 02110-1301, USA.
//----------------------------------------------------------------------------

#ifndef SVG_FONT_INCLUDED
#include "svg_element.h"
#include "svg_vertex_source.h"
#include "agg_font_cache_manager.h"
#define SVG_FONT_INCLUDED

namespace svg
{
typedef unsigned int unicode;

//////////////////////////////////////////////////////////////////////
// font: a helper element (non-existent in the source file) which is
//		 a vertex source for the glyph selected.
// It is a base class - its inheritors (font_win32, for example) are 
// used in the rest of the program
//////////////////////////////////////////////////////////////////////
class font : public element 
{
	public:
		enum weight_value
		{
			w100 = 100,
			w200 = 200,
			w300 = 300,
			normal = 400,
			w400 = 400,
			w500 = 500,
			w600 = 600,
			bold = 700,
			w700 = 700,
			w800 = 800,
			w900 = 900
		};

	public:
		//////////////////////////////////////////////////////////////////////
		// constructor and overriden virtual methods
		//////////////////////////////////////////////////////////////////////

		font(element *parent, unsigned level): element(parent, level) {}

		//Helper function: creates the font specification from separate
		//attribute values.
		//Arguments:
		//	buf:	a buffer for the specification text
		//	bufsize:buffer capacity
		//	family:	font family (for example: "Arial", "Times New Roman")
		//	style:	"italic" or "normal"
		//	weight: "100","200","300",.."900" or "normal", or "bold".
		//	parent:	an SVG graph element used to obtain information
		//			about any missing attribute, from the parent element's style
		static char *specification(	char *buf, size_t busize, char *family, 
									char *style, char *weight, element *parent);
		
		//////////////////////////////////////////////////////////////////////
		//Font specification is a text that describes, in the CSS, the font.
		//it contains up to four elements, separated by spaces:
		//	style:	"italic", "normal", or missing
		//	variant: "normal", or ""
		//	weight:	"bold", "normal", "100".."900", or missing
		//	size:	a size of the font (in fact - ignored, because we are using
		//			scalable glyph outlines);
		//	family: any name, that can contain the spaces.
		//
		//	new fonts are created from such specification. It seems to be unreasonable
		//  to build such text from dome source attibutes, then - in the initialize()
		//	function - to decompose it back to the particular fields in the font 
		//	description. There are two reasons:
		//	-	an unique name for the font, that fully describes its capabilities is 
		//		needed;
		//	-	beacuse the specification has its syntax that agree with the CSS
		//		font attribute - it is very easy to create the font, when it is 
		//		declared in such one attribute, instead of reading each of them
		//////////////////////////////////////////////////////////////////////
		
		//Initializes the font, using the font declaration string
		//Arguments:
		//	spec:	font specificaton
		void initialize(const char *spec);

		//prefix "ext:" is for elements that have no corresponding element in 
		//the XML SVG document structure
		virtual const char * tag() {return "ext:font";}; 

	public:
		//////////////////////////////////////////////////////////////////////
		// General information - to be implemented in the inherited class
		//////////////////////////////////////////////////////////////////////
		//Returns true, when the font is oblique
		virtual bool italic() = 0; 

		//Returns the weight of the font (from 100 to 900)
		virtual weight_value weight() = 0;

		//Returns the size (height of a capital letters) of the font
		virtual double size() = 0; 

		//Returns the name of the font family
		virtual const char *family() = 0;
		
		//////////////////////////////////////////////////////////////////////
		// Glyph: selection and geometry
		//////////////////////////////////////////////////////////////////////
		//Selects an actual glyph, which corresponds to the code
		//Arguments:
		//	code:			code of the glyph, that has to be selected
		//	apply_kerning:	true, if the kerning has to be used
		//Updates the m_code memeber!
		virtual void select(unicode code, bool apply_kerning = true) = 0;

		//Returns the actually selected character code
		unicode selected() {return m_code;}

		//Returns the vertex source, that contains selected's glyph outline:
		virtual vsource *vertices() = 0;

		//Returns the advance of the font direction
		virtual double advance_x() = 0;
		virtual double advance_y() = 0;
		//Returns a kernng values for current glyph, and the previous
		virtual double kerning_x() = 0;
		virtual double kerning_y() = 0;

		//Bounds of the glyph (rounded to the nearest integer)
		virtual const rectangle &bounds() = 0;

		//Actual glyph data (measured bounds, advances, etc.)
		//virtual const agg::glyph_cache *current() = 0;
	protected:
		//////////////////////////////////////////////////////////////////////
		// Other methods and fields
		//////////////////////////////////////////////////////////////////////
		//Called internally by the initialize(spec) function.
		// Arguments:
		//	family:	font family name, like "Arial" or "Times New Roman"
		//	italic: true, when the font has to in the italic version
		//	weight: weight of the font
		//
		//To be implemented in the inheritors classes.
		virtual void initialize(const char *family, bool italic, 
														weight_value weight) = 0;
		//Called internally to fill the m_spec buffer
		// Arguments:
		//	buf:	a buffer for the specification text
		//	bufsize:buffer capacity
		//	family:	font family name, like "Arial" or "Times New Roman"
		//	italic: true, when the font has to in the italic version
		//	weight: weight of the font
		char * specification(char *buf, size_t busize, const char *family, 
											bool italic, weight_value weight);

	protected:
		unicode m_code;						//actually selected character code
		char m_spec[SVG_MAX_ATTRIB_VALUE];	//font description (it is also its identifier)
}; //font

}; //namespace svg
#endif //SVG_FONT_INCLUDED

