//----------------------------------------------------------------------------
// SVG Viewer - Version 1.0
// A library of SVG renderer
// Copyright (C) 2008 Witold Jaworski
// Contact:	witold.jaworski@airplanes3d.net
//			http://airplanes3d.net
// 
// Contains elements from:
//			AGG - a graphic library (Copyright Maxim Shemanarev, mcseem@antigrain.com)
//          http://antigrain.com
// Uses functions from:
//          Expat - a XML parser library (Copyright James Clark)
// 
// SVG Viewer is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// SVG Viewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with SVG Viewer; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
// MA 02110-1301, USA.
//----------------------------------------------------------------------------

#ifndef SVG_REQ_RENDER_INCLUDED
#define SVG_REQ_RENDER_INCLUDED
#include "svg_basics.h"
#include "svg_request.h"
//#include "svg_styleset.h"
#include "svg_vertex_source.h"

#include "agg_conv_curve.h"
#include "agg_conv_contour.h"
#include "agg_conv_stroke.h"
#include "agg_conv_transform.h"
#include "agg_conv_dash.h"
#include "agg_vcgen_markers_term.h"
//
#include "agg_renderer_base.h"
#include "agg_renderer_scanline.h"
#include "agg_rasterizer_scanline_aa.h"
#include "agg_scanline_u.h"

#include "agg_span_allocator.h"

//////////////////////////////////////////////////////////////////////
// This module contains requests, that deals with the geometrical
// aspects of the SVG graph, or just its rendering into final image.
//
// These requests (req_measure, req_render, ...) are usually created 
// in the root (see svg_root.h) element of the SVG graph, which
// contains the objects required for request constructor: storages
// of the graph elements, their attributes and vertices. 
//////////////////////////////////////////////////////////////////////

namespace svg
{
//////////////////////////////////////////////////////////////////////
// reference: forward delclaration
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
// req_geometry: base class for the all geometry-based requests
// Contains not only SVG graph elements and attributes, like svg::request,
// but also the path storage and the global transformation
//////////////////////////////////////////////////////////////////////

struct req_geometry: public request
{
    public:

		//Constructor
		//Arguments:
		//	type:	type of the request
		//	paths:	path storage, that contains the element's paths
		//	attrs:	attribute storage, that contains the element's attributes
		//	elems:	element storage, that contains the SVG graph elements
		//	gtrans:	matrix, describing the initial transform
		req_geometry(kind type, path_storage &paths, attr_storage &attrs, 
									element_storage &elems, transform gtrans): 
		request(type, attrs,elems), 
		trans(gtrans),
		m_paths(paths), 
		m_vs(m_paths),
		m_curve(m_vs),
		m_c_curve(m_curve,m_trans),
		m_contour(m_c_curve),
		painter_trans(NULL)
		{
			expand(0.0);    //default outline width
		}

		//Simplified constructor (for get_xxx requests)
		//should receive the paths, attributes and storage
		//alter on. through an assign() call.
		//Argument:
		//	type:	type of this request.
		req_geometry(kind type) : request(type), 
		m_vs(m_paths),
		m_curve(m_vs),
		m_c_curve(m_curve,m_trans),
		m_contour(m_c_curve),
		painter_trans(NULL)
		{
			expand(0.0);    //default outline width
		}

		//And supplemental method that enables to place the 
		//information about the graph storage later on:
		//Arguments:
		//	paths:	path storage, that contains the element's paths
		//	attrs:	attribute storage, that contains the element's attributes
		//	elems:	element storage, that contains the SVG graph elements
		//	gtrans:	matrix, describing the initial transform
		void assign(path_storage &paths, attr_storage &attrs, 
							element_storage &elems, transform gtrans)
		{
			m_paths = vs_path(paths);
			m_vs.source(m_paths);
			trans = gtrans;
			request::assign(attrs,elems);
		}
		
		//expand property getter and setter:
		double expand() {return m_expand;}
		void expand(double e) 
		{
			m_expand = e; 
			m_contour.width(e);
		}

		//Standard way of processing a path, stored in the path_storage:
		//Arguments:
		//	path_id:	the id of the path, stored in the path_storage, 
		//				to be rendered
		//	ltrans:		local transformation matrix
		//	lstyle:		local style, that has to be used.
		virtual void perform(handle path_id, transform &ltrans, styleset &lstyle) = 0;

		//Alternate way of processing: with own vertex source
		//Arguments:
		//	vs:			a vertex source, with commands to draw the path
		//	ltrans:		local transformation matrix
		//	lstyle:		local style, that has to be used.
		//	path_id:	optional. the id of the path, stored in the vs, 
		//				to be rendered (there is no need for it, for the 
		//				text font glyphs, for example)
		void perform(vsource &vs, transform &ltrans, styleset &lstyle, 
														handle path_id = 0)
		{
			vsource &prev = m_vs.source(); //preserve orignal source...
			m_vs.source(vs); //...switch into the new one...
			perform(path_id, ltrans, lstyle); //... render something ...
			m_vs.source(prev); //...and restore to the original source!
		}

		//Returns reference to the default paths storage - useful for
		//svg::path element, when it has to return a vsource reference
		//for markers or a text on the path....
		vsource & paths()	{return m_paths;}

	public:
		styleset	style;	//a base (suggested) for modifications of the
							//local style, passed to the render() and measure method

		transform trans;	//base transformation - can be updated (and restored!) 
							//by a container, like a <g>
		transform *painter_trans; //helper - pointer to a transformation, that
								  //should be passed to the painter (fill or stroke)
								  //instead of global transform. Usually NULL, updated
								  //by the text, which uses diferent translations for
								  //each character.

		typedef agg::conv_curve<vertex_source>		curved;
		typedef agg::conv_transform<curved>			c_curved;
		typedef agg::conv_contour<c_curved>			contour; 

    protected:
		vs_path			m_paths;
		vertex_source	m_vs;
		transform		m_trans; //actual transformation (for the path_trans)

		curved			m_curve;
		c_curved		m_c_curve;
		contour			m_contour;

		double			m_expand;	//offset of the shape contour (usually = 0)

}; //req_geometry

//////////////////////////////////////////////////////////////////////
// req_render: renders the element and its descendants into final image.
// Used by SVG graph root node to produce the result.
//////////////////////////////////////////////////////////////////////

struct req_render: public req_geometry
{
	public:

	//flags, set in the filter buffer (svg::source_graphic_buffer)
	//to signalize a tip to the rendered shape(s):
    enum flag
    {
			none				= 0, //empty value
			no_stroke			= 16 //do not use stroke (render just filled contours, only)	
									 //(used to render masks - for the clipping)
	};

	public:

		//Constructor
		//Arguments:
		//	buffer: buffer in the memory - target of this rendering
		//	clip:	cliping rectangle (AFTER transformation through gtrans)
		//	bkgnd:	background color (with alfa!)
		//	paths:	path storage, that contains the element's paths
		//	attrs:	attribute storage, that contains the element's attributes
		//	elems:	element storage, that contains the SVG graph elements
		//	gtrans:	matrix, describing the initial transform
		req_render (pixfmt &buffer,  rectangle clip, rgba8 bkgnd, 
					path_storage &paths, attr_storage &attrs, 
					element_storage &elems, transform &gtrans):
					req_geometry(request::render, paths,attrs,elems,gtrans), 
		m_dash(m_curve),
		m_dash_stroke(m_dash),
		m_dash_outline(m_dash_stroke, m_trans),
		m_solid_stroke(m_curve),
		m_solid_outline(m_solid_stroke, m_trans),
		m_pixf(buffer),
		m_rnb(m_pixf),
		m_rns(m_rnb),
		bounds(clip),
		m_flag(0)
		{
			m_rnb.clear(bkgnd);
			m_ras.clip_box(clip.x1, clip.y1, clip.x2, clip.y2);
			gamma(1.5);		//default gamma correction
		} 	

		//Simplified constructor (for get_render contained requests)
		//Arguments:
		//	buffer: buffer in the memory - target of this rendering
		//	clip:	cliping rectangle (AFTER transformation through gtrans)
		//	bkgnd:	background color (with alfa!)
		//
		//Requires additional call to req_geometry::assign(), to be usable 
		req_render(pixfmt &buffer,  rectangle clip, rgba8 bkgnd) : 
			req_geometry(request::render),
			m_dash(m_curve),
			m_dash_stroke(m_dash),
			m_dash_outline(m_dash_stroke, m_trans),
			m_solid_stroke(m_curve),
			m_solid_outline(m_solid_stroke, m_trans),
			m_pixf(buffer),
			m_rnb(m_pixf),
			m_rns(m_rnb),
			m_bkgnd(bkgnd),
			bounds(clip),
			restricted_filter(invalid_handle),
			m_flag(none)
		{
			m_rnb.clear(bkgnd);
			m_ras.clip_box(clip.x1, clip.y1, clip.x2, clip.y2);
			gamma(1.5);		//default gamma correction
		}

	public: //typedefs 
		//first case of vertex pipeline: solid
		typedef agg::conv_stroke<curved>			solid_stroked;
		typedef agg::conv_transform<solid_stroked>	solid_outlined;
		//second case of vertex pipeline: dashed
		typedef agg::conv_dash<curved>				dashed;
		typedef agg::conv_stroke<dashed>			dash_stroked;
		typedef agg::conv_transform<dash_stroked>	dash_outlined;
		//("curved" type - see typedefs in svg::req_geometry, above)

        typedef agg::renderer_base<pixfmt>					renderer;
        typedef agg::renderer_scanline_aa_solid<renderer>	renderer_solid;
		typedef agg::rasterizer_scanline_aa<>				rasterizer;
		typedef agg::scanline_u8							scanline;

	public:	//public fields - can be modified (trans is defined in the base class)
		rectangle	bounds;	//actual global clip box.
		
		handle		restricted_filter; //this filter should not be used!
									   //(needed in the svg:source_graphic_buffer)

	public:
		//Standard way of rendering a path, stored in the path_storage:
		//Arguments:
		//	path_id:	the id of the path, stored in the path_storage, 
		//				to be rendered
		//	ltrans:		local transformation matrix
		//	lstyle:		local style, that has to be used.
		virtual void perform(handle path_id, transform &ltrans, styleset &lstyle);

		//Returns the renderer_base object - can be used in less typical
		//algorithms for redering some objects
		renderer &renb() {return m_rnb;}

		//background color - for eventual filter buffers
		rgba8 background() {return m_bkgnd;}

		//gamma property getter and setter:
		double gamma() {return m_gamma;}
		void gamma(double g) 
		{
			m_gamma = g; 
			m_ras.gamma(agg::gamma_power(g));
		}


		//flag operations:
		void set(flag f) {m_flag = (m_flag | f);}
		void clear(flag f) {m_flag = (m_flag & ~f);}
		bool has(flag f) { return ((m_flag & f) != 0);}
	
    protected:
		double			m_gamma;		//anti-aliasing factor

		dashed			m_dash;
		dash_stroked	m_dash_stroke;	//outline without transformation
		dash_outlined	m_dash_outline;	

		solid_stroked	m_solid_stroke;	//outline without transformation
		solid_outlined	m_solid_outline;	

		pixfmt			m_pixf;
		renderer		m_rnb;
		renderer_solid	m_rns;

		rgba8			m_bkgnd;	//background color - for the filters
		
		//utility objects - used just for the default rendering 
		//(to not create them for each rendered element)
		rasterizer					m_ras;
		scanline					m_scl;
		agg::span_allocator<rgba8>	m_allocator;

		//utility fileds - for rendering to the filter buffer:
		unsigned		m_flag; //various req_render::flags
}; //req_render


//////////////////////////////////////////////////////////////////////
// req_measure: summary information about the element and its descendants
// Used by SVG graph root node to gather some general information 
// about its content (overall size, number of vertices, etc...)
//////////////////////////////////////////////////////////////////////

// Helper class: this is an "overlay" over a normal vertex source - 
// exists for the statistics, only: just to count the vertices
template<class VertexSource> class conv_count 
{
	public:
		conv_count(VertexSource& vs) : m_source(&vs), m_count(0) {}

        void count(unsigned n) { m_count = n; }
        unsigned count() const { return m_count; }

        void rewind(unsigned path_id) { m_source->rewind(path_id); }
        unsigned vertex(double* x, double* y) 
        { 
            ++m_count; 
            return m_source->vertex(x, y); 
        }

    private:
        VertexSource* m_source;
        unsigned m_count;
};

struct req_measure: public req_geometry
{
    public:
		//Constructor
		//Arguments:
		//	paths:	path storage, that contains the element's paths
		//	attrs:	attribute storage, that contains the element's attributes
		//	elems:	element storage, that contains the SVG graph elements
		//	gtrans:	matrix, describing the initial transform
		req_measure(path_storage &paths, attr_storage &attrs, 
									element_storage &elems, transform &gtrans): 
		req_geometry(request::measure, paths,attrs,elems,gtrans), 
		m_curved_count(m_c_curve),
		bounds(0,0,-1,-1) //is_valid() returns false
		{	
			vertex_count = 0;
		} 	

		//Simplified constructor (for get_render contained requests)
		//Requires additional call to req_geometry::assign(), to be usable 
		req_measure(): 
		req_geometry(request::measure), 
		m_curved_count(m_c_curve),
		bounds(0,0,-1,-1) //is_valid() returns false
		{	
			vertex_count = 0;
		} 	

	public: //public field - can be modified
		rectangle	bounds;		//real size of the SVG graph
								//nominal size (delcared in the first <svg>)
		double width;			//width of the "paper"
		double height;			//height of the "paper" 
		unsigned vertex_count;	//number of the vertices

	public:
		//Standard way of measuring a path, stored in the path_storage:
		//Arguments:
		//	path_id:	the id of the path, stored in the path_storage, 
		//				to be rendered
		//	ltrans:		local transformation matrix
		//	lstyle:		local style, that has to be used.
		virtual void perform(handle path_id, transform &ltrans, styleset &lstyle);

    protected:
		typedef conv_count<c_curved> curved_count;
		curved_count	m_curved_count;
}; //req_measure

}; //namespace svg
#endif //SVG_REQ_RENDER_INCLUDED

