//----------------------------------------------------------------------------
// SVG Viewer - Version 1.0
// A library of SVG renderer
// Copyright (C) 2008 Witold Jaworski
// Contact:	witold.jaworski@airplanes3d.net
//			http://airplanes3d.net
// 
// Contains elements from:
//			AGG - a graphic library (Copyright Maxim Shemanarev, mcseem@antigrain.com)
//          http://antigrain.com
// Uses functions from:
//          Expat - a XML parser library (Copyright James Clark)
// 
// SVG Viewer is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// SVG Viewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with SVG Viewer; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
// MA 02110-1301, USA.
//----------------------------------------------------------------------------
#include "svg_exception.h"
#include "svg_tspan.h"

namespace svg
{
	//Helper function. returns true, when a character is utf-8 encoded
	//(this means its bits are 10xxxxxx)
	//Arguments:
	//	c:	a character to be checked
	//Returns true, when character value bits begins with 10xxxxxx.
	//(All "value" utf-8 bits have such bit pattern)
	static inline bool is_encoded(char c) {return (c & 0xC0) == 0x80;}
//----------------------------------------------------------------------------
	//Helper function. Advances the given chcaracter string by a number of
	//characters that describe one unicode value (it can be 1, 2, 3, or 4
	//characters - see utf-8 specification).
	//Arguments:
	//	t:	ponter to a character array. It is increased during this function
	//Returns: an unicode character (32-bit)
	static unicode decode_utf8(char **t)
	{
		unicode c = 0; //encoded sequence
		unsigned char *p = (unsigned char *) *t;
		if (p[0] < 0x80) //not encoded
		{
			c = p[0]; 
			(*t) += 1; //advance to the next character
		}
		else
		if (p[0] < 0xE0 && is_encoded(p[1]))
		{
			c = (p[0] & 0x1F) << 6; 
			c|= (p[1] & 0x3F);
			(*t) += 2;
		}
		else
		if (p[0] < 0xF0 && is_encoded(p[1]) && is_encoded(p[2]))
		{
			c = (p[0] & 0x0F) << 12;
			c|= (p[1] & 0x3F) << 6;
			c|= (p[2] & 0x3F);
			(*t) += 3;
		}
		else
		if (p[0] < 0xF8 && is_encoded(p[1]) && is_encoded(p[2]) && is_encoded(p[3]))
		{
			c = (p[0] & 0x0F) << 18;
			c|= (p[1] & 0x3F) << 12;
			c|= (p[2] & 0x3F) << 6;
			c|= (p[3] & 0x3F);
			(*t) += 4;
		}

		return c;
	}
//----------------------------------------------------------------------------
	//Appends another fragment of a text to the text already accumulated
	//Arguments:
	//	utf8:	text content, (encoded as utf-8) - may not be null-terminated
	//	len:	number of characters to analyze and convert into unicode
	void tspan::append(const char *utf8, unsigned len)
	{
		char *p = (char *) utf8;
		while(p < (utf8 + len))
		{
			unicode c = decode_utf8(&p);
			if (c >= ' ') m_text.add(c); //control characters (< ' ') are skipped
		}
	}
//----------------------------------------------------------------------------
	//Renders or measures the content of a single <tspan>. When this instance 
	//contains its own x,y atributes, which enables it to fix the text position, 
	//or the text position is already set up by a previous tspan, the text will
	//be measured/rendered.
	//Returns the id of the next element
	handle tspan::render(req_geometry &req, text_position &tpos)
	{
		styleset	style	= req.style;
		style.update(req,m_attribute_first, m_attribute_last);
		tpos.apply(req,style);
			
		int fcount;
		if (m_fcount > m_text.size())  //this may happen - more constrains than characters!
		{
			fcount = m_text.size() - 1;
		}
		else
		{
			fcount = m_fcount - 1;
		}

		int i = 0;
		while (i <= fcount || i == 0) //i == 0: initial run should pass, 
									 //even when fcount < 0!
		{
			unsigned j;
			int count = i < fcount? 1: m_text.size() - fcount;

			//if the tpos has the font loaded properly, after this call to its set()
			//method it should be "valid":
			tpos.set(x(req,i,tpos.x()),y(req,i,tpos.y()),m_text,i, (unsigned) count,
										dx(req,i,0.0),dy(req,i,0.0),rot(req,i,0.0));

			for (j = 0; j < tpos.count(); j++) //tpos.count() may be < count!
			{

				if (tpos.is_valid()) //i.e. is somewhere on the path
				{
					transform t = tpos.transformation(); //initial transformation of the
														 //glyph
					t *= m_trans;
					req.perform(*tpos.outline(),t,style); //inside: ultimate transform = 
				}										  //t * req.trans
				
				tpos.advance();
			}//for
		
			i += count;
		}//while

		return skip(req);
	}

} // namespace svg
